:- lib(ic).
:- lib(ic_global).
:- lib(branch_and_bound).
:- lib(util).
:- lib(hash).

% too lazy...
:-dynamic(allowed/2).

%stack(+Ls,-Lp) is true if
%Ls is a list of N boxes
%Lp is a list of P stacks where the maximum height of the stacks is N/P
stack(Ls, List):-
        length(Ls, N),
        Select = first_fail,
        Choice = indomain_min,
        Method = complete,
        stack(Ls, N, List, Select, Choice, Method),
        !.
stack(_,_).

stack(Ls, N, Stacks, Select, Choice, Method):-
        retractall(allowed(_I1,_I2)),
        length(Ls,Nb),
        set_stack_heights(N, StackList),
        flatten(StackList, List),
        length(List, NbAll),
        NbFakeBoxes is NbAll - Nb,
        fill(Ls, NbFakeBoxes, La),
        % we use classes of boxes
        % this should be more efficient if we have duplicates
        build_classes(La, Cs, NCs),
        % we label classes of boxes
        List #:: [1..NCs],
        % set constraints
        set_occurrences(Cs, List),
        compute_allowed(Cs),
        set_stack_constraint(StackList),
        set_stack_cost(StackList, NCs, Cost),
        Cost #:: 0..N,
        set_stack_height_constraint(NCs, N, Cost, StackList),
        empty_stacks(NCs, StackList),
        % search...
        bb_min(
                  search(List, 0, Select, Choice, Method, []),
                  Cost,
                  StackList,
                  StackList,
                  _Opt,
                  bb_options{strategy:restart, timeout:60}
              ),
        generate_solution(Cs, StackList, Stacks).


% expand classes, ignore dummy boxes
generate_solution(Classes, StackList, Stacks) :-
        (
            foreach(Stack, StackList),
            foreach(ExpStack, Stacks),
            param(Classes)
        do
            (
                foreach(C, Stack),
                fromto([], I, O, ExpStackR),
                param(Classes)
            do
                once(member(class(C, W, L, _), Classes)),
                (
                    W == 0
                ->
                    % dummy box: ignore
                    O = I
                ;
                    O = [(W,L)|I]
                )
            ),
            reverse(ExpStackR, ExpStack)
        ).

% we can model the N stacks with different heights:
% the later stacks can't have full height, so that equation H=NB/NS
% is fulfilled
set_stack_heights(N, StackList) :-
        (
            for(I, 1, N),
            foreach(Stack, StackList),
            param(N)
        do
            Tmp is N/I,
            ceiling(Tmp,T2),
            integer(T2,H),
            length(Stack, H)
        ).


% turn list of boxes into a list of box classes
% boxes are sorted descendingly by width
% this will ensure that dummy boxes will get largest class index
% also, when labelling using indomain_min, we try widest boxes first
build_classes(La, ClassList, NCs) :-
        hash_create(Cs),
        sort(1, >=, La, SortedLa),
        build_classes0(SortedLa, 0, Cs, NCs),
        hash_list(Cs, _, ClassList).

build_classes0([], NCs, _, NCs).
build_classes0([box(W,L)|R], I, Cs, NCs) :-
        (
            hash_contains(Cs, box(W,L))
        ->
            hash_get(Cs, box(W,L), class(C, W, L, N)),
            N1 is N+1,
            hash_set(Cs, box(W,L), class(C, W, L, N1)),
            build_classes0(R, I, Cs, NCs)
        ;
            I1 is I+1,
            hash_set(Cs, box(W,L), class(I1, W, L, 1)),
            build_classes0(R, I1, Cs, NCs)
        ).

% each class C must occur exactly N times in the stacks
set_occurrences(Cs, PIS) :-
        (
            foreach(class(C, _, _, N), Cs),
            param(PIS)
        do
            ic_global:occurrences(C, PIS, N)
        ).

% compute that facts that state which box classes can be put
%  on top of a given box class
compute_allowed(Cs) :-
        (
            foreach(class(I1, W1, L1, _), Cs),
            param(Cs)
        do
            findall(I2,
                    (
                        member(class(I2, W2, L2, _), Cs),
                        W2 =< W1,
                        L2 =< L1
                    ),
                    I2s),
            assert(allowed(I1, I2s))
        ).


% set stack constraint:
% once a box is known, compute the domain of box classes that can be put on top
% and restrict the domain of the box above accordingly
set_stack_constraint(StackList) :-
        (
            foreach(Stack, StackList)
        do
            set_stack_constraint0(Stack)
        ).

set_stack_constraint0([_]) :- !.
set_stack_constraint0([H | T]) :-
        T = [T1 | _],
        suspend(sc(H,T1,Susp), 4, [H->constrained,H->inst], Susp),
        set_stack_constraint0(T).

:- demon sc/3.
sc(H, T, Susp) :-
        get_domain_as_list(H, HD),
        (
            foreach(I1, HD),
            foreach(I2s, TDs)
        do
            allowed(I1, I2s)
        ),
        flatten(TDs, TD),
        T #:: TD,
        (
            ground(H)
        ->
            kill_suspension(Susp)
        ;
            true
        ).


% each stack has unit cost
% we can determine whether a stack is empty by looking at the "bottom" variable
% if the variable is bound to the class index of the dummy boxes (= number of classes),
% then the stack is empty.
% On the other hand, if the cost value for the stack is zero, then the bottom box
% must belong to the dummy box class
set_stack_cost(StackList, NCs, Cost) :-
        (
            foreach([F|_], StackList),
            foreach(C, Sum),
            param(NCs)
        do
            C #:: 0..1,
            ( F #\= NCs ) => ( C #= 1 ),
            ( F #=  NCs ) => ( C #= 0 ),
            ( C #= 0 ) => ( F #=  NCs ),
            ( C #= 1 ) => ( F #\= NCs )
        ),
        ic_global:sumlist(Sum,Cost).


% if we know that one stack is empty, then all the stacks to the right
% must also be empty
empty_stacks(NCs, StackList) :-
        (
            fromto(StackList, [[F|_]|Rest], Rest, []),
            param(NCs)
        do
            suspend(es(F, NCs, Rest), 4, [F->inst])
        ).

es(F, NCs, Rest) :-
        (
            F == NCs
        ->
            (
                foreach([F1|_],Rest),
                param(NCs)
            do
                F1 #= NCs
            )
        ;
            true
        ).


% we can set two constraints regarding the stack height...
set_stack_height_constraint(NCs, N, Cost, StackList) :-
        set_stack_height_constraint1(NCs, N, Cost, StackList),
        set_stack_height_constraint2(NCs, N, Cost, StackList).

% first stack height constraint:
% if the Cost is constrained, then we can calculate the maximum stack height
% any boxes above that height must belong to the dummy box class
set_stack_height_constraint1(NCs, N, Cost, StackList) :-
        H $= N/Cost,
        (
            foreach(Stack, StackList),
            param(H, NCs)
        do
            length(Stack, L),
            StackTerm =.. [s|Stack],
            suspend(shc(H, L, StackTerm, NCs, Susp), 4, [H->constrained], Susp)
        ).

:- demon shc/5.
shc(H, L, StackR, NCs, Susp) :-
        get_min(H, Min),
        get_max(H, Max),
        (
            Min > L
        ->
            kill_suspension(Susp)
        ;
            Max < L
        ->
            A is integer(ceiling(Max))+1,
            (
                A > L
            ->
                true
            ;
                arg(A, StackR, F),
                F #= NCs
            ),
            ( ground(H) -> kill_suspension(Susp) ; true )
        ;
            true
        ).

% second stack height constraint
% when the height of a stack becomes known
% (after placing the first box in the stack that belongs to the dummy class)
% we can determine the minimum number of empty stacks that must follow
set_stack_height_constraint2(NCs, N, Cost, StackList) :-
        SLT =.. [slt|StackList],
        (
            foreach(Stack, StackList),
            param(N, Cost, NCs, SLT)
        do
            Stack = [F|Rest],
            (
                Rest == []
            ->
                true
            ;
                (
                    foreach(F2, Rest),
                    fromto(F, F1, F2, Last),
                    count(Height, 1, StackHeight),
                    param(N, Cost, NCs, SLT)
                do
                    % when height is known, we can determine max number of stacks
                    NStacks is integer(ceiling(N/Height)),
                    suspend(shc2(F1, F2, NCs, NStacks, Cost, N, SLT, Susp), 4, [F1->inst,F2->inst], Susp)
                ),
                NStacks is integer(ceiling(N/StackHeight)),
                suspend(shc2(Last, NCs, NCs, NStacks, Cost, N, SLT, Susp), 4, [Last->inst], Susp)
            )
        ).

:- demon shc2/8.
shc2(F1, F2, NCs, NStacks, Cost, N, SLT, Susp) :-
        (
            ground(F1),
            ground(F2),
            F1 \== NCs,
            F2  == NCs
        ->
            Cost #=< NStacks,
            (
                for(I, NStacks+1, N),
                param(SLT, NCs)
            do
                arg(I, SLT, [SF|_]),
                SF #= NCs
            ),
            kill_suspension(Susp)
        ;
            true
        ).

%fill(+L,+I,-Ls) is true if
%L is a list of boxes
%I is a number >= 0
%Ls is the list L in wich we added I fake boxes at the beginning of he list
fill(L,0,L):-!.
fill(L,I,[box(0,0)|R]):-I>0, I1 is I-1, fill(L,I1,R).
